#ifndef MATRIX_H_
#define MATRIX_H_

#include "Vector.h"
using namespace arda::Vector;

#include <cmath>
#include <string>

namespace arda 
   {
   namespace Matrix
      {
      //////////////////////////////////////////////////////////////////////////
      // Maxtrices
      //
      // Supported operations on matrices
      //
      // getstring*()	Returns a printable string representation of the matrix.
      // assign*()	Assign the value of one matrix to another.
      // add*()		Add two matrices
      // subtract*()	A convenience function to avoid writing 
      //			add(m1, scale(m2, -1))
      // transpose*()	Returns the transpose of a matrix.
      // scale*()	Multiplies a matrix by a scalar.
      // multvec*()	Multiplies a matrix by a vector.
      // multmat*()	Multiplies a matrix by another matrix.
      // det*()		Calculates the determinant of a matrix.
      // inverse*()	Returns the inverse of a non-singular matrix.
      // setidentity*()	Sets a matrix to the identity matrix.
      // getcol*()	Returns the vector that is column n of the matrix.
      // setcol*()	Set the vector that is column n of the matrix.
      //
      // assign*(), add*(), and subtract*() are defined inline, though this might
      // might not be the best thing.  Some real world testing is needed to decide
      // whether or not htye should be remain inline.
      
      typedef int Matrix2i[4];
      typedef float Matrix2f[4];
      typedef double Matrix2d[4];
      
      typedef int Matrix3i[9];
      typedef float Matrix3f[9];
      typedef double Matrix3d[9];
      
      typedef int Matrix4i[16];
      typedef float Matrix4f[16];
      typedef double Matrix4d[16];
      
      //////////////////////////////////////////////////////////////////////////
      // getstring*()
      std::string & getstring2m(Matrix2i m, std::string & s);
      std::string & getstring2m(Matrix2f m, std::string & s);
      std::string & getstring2m(Matrix2d m, std::string & s);
      
      std::string & getstring3m(Matrix3i m, std::string & s);
      std::string & getstring3m(Matrix3f m, std::string & s);
      std::string & getstring3m(Matrix3d m, std::string & s);

      std::string & getstring4m(Matrix4i m, std::string & s);
      std::string & getstring4m(Matrix4f m, std::string & s);
      std::string & getstring4m(Matrix4d m, std::string & s);

      //////////////////////////////////////////////////////////////////////////
      // assign*()

      inline int * assign2m(Matrix2i m1, Matrix2i m2)
 	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; 
	 m1[2] = m2[2]; m1[3] = m2[3]; 
	 return m1; 
	 }
      inline float * assign2m(Matrix2f m1, Matrix2f m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; 
	 m1[2] = m2[2]; m1[3] = m2[3]; 
	 return m1; 
	 }
      inline double * assign2m(Matrix2d m1, Matrix2d m2)
	 {
	 m1[0] = m2[0]; m1[1] = m2[1]; 
	 m1[2] = m2[2]; m1[3] = m2[3]; 
	 return m1; 
	 }

      inline int * assign3m(Matrix3i m1, Matrix3i m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; m1[2] = m2[2];
	 m1[3] = m2[3]; m1[4] = m2[4]; m1[5] = m2[5]; 	
	 m1[6] = m2[6]; m1[7] = m2[7]; m1[8] = m2[8]; 
	 return m1; 
	 }
      inline float * assign3m(Matrix3f m1, Matrix3f m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; m1[2] = m2[2];
	 m1[3] = m2[3]; m1[4] = m2[4]; m1[5] = m2[5]; 	
	 m1[6] = m2[6]; m1[7] = m2[7]; m1[8] = m2[8]; 
	 return m1; 
	 }
      inline double * assign3m(Matrix3d m1, Matrix3d m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; m1[2] = m2[2];
	 m1[3] = m2[3]; m1[4] = m2[4]; m1[5] = m2[5]; 	
	 m1[6] = m2[6]; m1[7] = m2[7]; m1[8] = m2[8]; 
	 return m1; 
	 }

      inline int * assign4m(Matrix4i m1, Matrix4i m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; m1[2] = m2[2]; m1[3] = m2[3];
	 m1[4] = m2[4]; m1[5] = m2[5]; m1[6] = m2[6]; m1[7] = m2[7];
	 m1[8] = m2[8]; m1[9] = m2[9]; m1[10] = m2[10]; m1[11] = m2[11];
	 m1[12] = m2[12]; m1[13] = m2[13]; m1[14] = m2[14]; m1[15] = m2[15];
	 return m1; 
	 }
      inline float * assign4m(Matrix4f m1, Matrix4f m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; m1[2] = m2[2]; m1[3] = m2[3];
	 m1[4] = m2[4]; m1[5] = m2[5]; m1[6] = m2[6]; m1[7] = m2[7];
	 m1[8] = m2[8]; m1[9] = m2[9]; m1[10] = m2[10]; m1[11] = m2[11];
	 m1[12] = m2[12]; m1[13] = m2[13]; m1[14] = m2[14]; m1[15] = m2[15];
	 return m1; 
	 }
      inline double * assign4m(Matrix4d m1, Matrix4d m2)
	 { 
	 m1[0] = m2[0]; m1[1] = m2[1]; m1[2] = m2[2]; m1[3] = m2[3];
	 m1[4] = m2[4]; m1[5] = m2[5]; m1[6] = m2[6]; m1[7] = m2[7];
	 m1[8] = m2[8]; m1[9] = m2[9]; m1[10] = m2[10]; m1[11] = m2[11];
	 m1[12] = m2[12]; m1[13] = m2[13]; m1[14] = m2[14]; m1[15] = m2[15];
	 return m1; 
	 }

      //////////////////////////////////////////////////////////////////////////
      // add*()

      inline int * add2m(Matrix2i m1, Matrix2i m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1];
	 m1[2] += m2[2]; m1[3] += m2[3];
	 return m1;
	 }
      inline float * add2m(Matrix2f m1, Matrix2f m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1];
	 m1[2] += m2[2]; m1[3] += m2[3];
	 return m1;
	 }
      inline double * add2m(Matrix2d m1, Matrix2d m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1];
	 m1[2] += m2[2]; m1[3] += m2[3];
	 return m1;
	 }

      inline int * add3m(Matrix3i m1, Matrix3i m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1]; m1[2] += m2[2]; 
	 m1[3] += m2[3]; m1[4] += m2[4]; m1[5] += m2[5];
	 m1[6] += m2[6]; m1[7] += m2[7]; m1[8] += m2[8];
	 return m1;
	 }
      inline float * add3m(Matrix3f m1, Matrix3f m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1]; m1[2] += m2[2]; 
	 m1[3] += m2[3]; m1[4] += m2[4]; m1[5] += m2[5];
	 m1[6] += m2[6]; m1[7] += m2[7]; m1[8] += m2[8];
	 return m1;
	 }
      inline double * add3m(Matrix3d m1, Matrix3d m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1]; m1[2] += m2[2]; 
	 m1[3] += m2[3]; m1[4] += m2[4]; m1[5] += m2[5];
	 m1[6] += m2[6]; m1[7] += m2[7]; m1[8] += m2[8];
	 return m1;
	 }

      inline int * add4m(Matrix4i m1, Matrix4i m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1]; m1[2] += m2[2]; m1[3] += m2[3]; 
	 m1[4] += m2[4]; m1[5] += m2[5]; m1[6] += m2[6]; m1[7] += m2[7]; 
	 m1[8] += m2[8]; m1[9] += m2[9]; m1[10] += m2[10]; m1[11] += m2[11];
	 m1[12] += m2[12]; m1[13] += m2[13]; m1[14] += m2[14]; m1[15] += m2[15];
	 return m1;
	 }
      inline float * add4m(Matrix4f m1, Matrix4f m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1]; m1[2] += m2[2]; m1[3] += m2[3]; 
	 m1[4] += m2[4]; m1[5] += m2[5]; m1[6] += m2[6]; m1[7] += m2[7]; 
	 m1[8] += m2[8]; m1[9] += m2[9]; m1[10] += m2[10]; m1[11] += m2[11];
	 m1[12] += m2[12]; m1[13] += m2[13]; m1[14] += m2[14]; m1[15] += m2[15];
	 return m1;
	 }
      inline double * add4m(Matrix4d m1, Matrix4d m2)
	 {
	 m1[0] += m2[0]; m1[1] += m2[1]; m1[2] += m2[2]; m1[3] += m2[3]; 
	 m1[4] += m2[4]; m1[5] += m2[5]; m1[6] += m2[6]; m1[7] += m2[7]; 
	 m1[8] += m2[8]; m1[9] += m2[9]; m1[10] += m2[10]; m1[11] += m2[11];
	 m1[12] += m2[12]; m1[13] += m2[13]; m1[14] += m2[14]; m1[15] += m2[15];
	 return m1;
	 }

      //////////////////////////////////////////////////////////////////////////
      // subtract*()

      inline int * subtract2m(Matrix2i m1, Matrix2i m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1];
	 m1[2] -= m2[2]; m1[3] -= m2[3];
	 return m1;
	 }
      inline float * subtract2m(Matrix2f m1, Matrix2f m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1];
	 m1[2] -= m2[2]; m1[3] -= m2[3];
	 return m1;
	 }
      inline double * subtract2m(Matrix2d m1, Matrix2d m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1];
	 m1[2] -= m2[2]; m1[3] -= m2[3];
	 return m1;
	 }

      inline int * subtract3m(Matrix3i m1, Matrix3i m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1]; m1[2] -= m2[2]; 
	 m1[3] -= m2[3]; m1[4] -= m2[4]; m1[5] -= m2[5];
	 m1[6] -= m2[6]; m1[7] -= m2[7]; m1[8] -= m2[8];
	 return m1;
	 }
      inline float * subtract3m(Matrix3f m1, Matrix3f m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1]; m1[2] -= m2[2]; 
	 m1[3] -= m2[3]; m1[4] -= m2[4]; m1[5] -= m2[5];
	 m1[6] -= m2[6]; m1[7] -= m2[7]; m1[8] -= m2[8];
	 return m1;
	 }
      inline double * subtract3m(Matrix3d m1, Matrix3d m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1]; m1[2] -= m2[2]; 
	 m1[3] -= m2[3]; m1[4] -= m2[4]; m1[5] -= m2[5];
	 m1[6] -= m2[6]; m1[7] -= m2[7]; m1[8] -= m2[8];
	 return m1;
	 }

      inline int * subtract4m(Matrix4i m1, Matrix4i m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1]; m1[2] -= m2[2]; m1[3] -= m2[3]; 
	 m1[4] -= m2[4]; m1[5] -= m2[5]; m1[6] -= m2[6]; m1[7] -= m2[7]; 
	 m1[8] -= m2[8]; m1[9] -= m2[9]; m1[10] -= m2[10]; m1[11] -= m2[11];
	 m1[12] -= m2[12]; m1[13] -= m2[13]; m1[14] -= m2[14]; m1[15] -= m2[15];
	 return m1;
	 }
      inline float * subtract4m(Matrix4f m1, Matrix4f m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1]; m1[2] -= m2[2]; m1[3] -= m2[3]; 
	 m1[4] -= m2[4]; m1[5] -= m2[5]; m1[6] -= m2[6]; m1[7] -= m2[7]; 
	 m1[8] -= m2[8]; m1[9] -= m2[9]; m1[10] -= m2[10]; m1[11] -= m2[11];
	 m1[12] -= m2[12]; m1[13] -= m2[13]; m1[14] -= m2[14]; m1[15] -= m2[15];
	 return m1;
	 }
      inline double * subtract4m(Matrix4d m1, Matrix4d m2)
	 {
	 m1[0] -= m2[0]; m1[1] -= m2[1]; m1[2] -= m2[2]; m1[3] -= m2[3]; 
	 m1[4] -= m2[4]; m1[5] -= m2[5]; m1[6] -= m2[6]; m1[7] -= m2[7]; 
	 m1[8] -= m2[8]; m1[9] -= m2[9]; m1[10] -= m2[10]; m1[11] -= m2[11];
	 m1[12] -= m2[12]; m1[13] -= m2[13]; m1[14] -= m2[14]; m1[15] -= m2[15];
	 return m1;
	 }

      //////////////////////////////////////////////////////////////////////////
      // scale*()

      inline int * scale2m(Matrix2i m, int s)
	 { 
	 m[0] *= s; m[1] *= s; 
	 m[2] *= s; m[3] *= s; 
	 return m;
	 }
      inline float * scale2m(Matrix2f m, float s)
	 { 
	 m[0] *= s; m[1] *= s; 
	 m[2] *= s; m[3] *= s; 
	 return m;
	 }
      inline double * scale2m(Matrix2d m, double s)
	 { 
	 m[0] *= s; m[1] *= s; 
	 m[2] *= s; m[3] *= s; 
	 return m;
	 }

      inline int * scale3m(Matrix3i m, int s)
	 { 
	 m[0] *= s; m[1] *= s; m[2] *= s;
	 m[3] *= s; m[4] *= s; m[5] *= s;
	 m[4] *= s; m[5] *= s; m[6] *= s;
	 return m;
	 }
      inline float * scale3m(Matrix3f m, float s)
	 { 
	 m[0] *= s; m[1] *= s; m[2] *= s;
	 m[3] *= s; m[4] *= s; m[5] *= s;
	 m[4] *= s; m[5] *= s; m[6] *= s;
	 return m;
	 }
      inline double * scale3m(Matrix3d m, double s)
	 { 
	 m[0] *= s; m[1] *= s; m[2] *= s;
	 m[3] *= s; m[4] *= s; m[5] *= s;
	 m[4] *= s; m[5] *= s; m[6] *= s;
	 return m;
	 }

      inline int * scale4m(Matrix4i m, int s)
	 { 
	 m[0] *= s; m[1] *= s; m[2] *= s; m[3] *= s; 
	 m[4] *= s; m[5] *= s; m[4] *= s; m[5] *= s;
	 m[6] *= s; m[7] *= s; m[8] *= s; m[9] *= s;
	 m[10] *= s; m[11] *= s; m[12] *= s; m[13] *= s;
	 return m;
	 }
      inline float * scale4m(Matrix4f m, float s)
	 { 
	 m[0] *= s; m[1] *= s; m[2] *= s; m[3] *= s; 
	 m[4] *= s; m[5] *= s; m[4] *= s; m[5] *= s;
	 m[6] *= s; m[7] *= s; m[8] *= s; m[9] *= s;
	 m[10] *= s; m[11] *= s; m[12] *= s; m[13] *= s;
	 return m;
	 }
      inline double * scale4m(Matrix4d m, double s)
	 { 
	 m[0] *= s; m[1] *= s; m[2] *= s; m[3] *= s; 
	 m[4] *= s; m[5] *= s; m[6] *= s; m[7] *= s;
	 m[8] *= s; m[9] *= s; m[10] *= s; m[11] *= s;
	 m[12] *= s; m[13] *= s; m[14] *= s; m[15] *= s;
	 return m;
	 }

      //////////////////////////////////////////////////////////////////////////
      // multvec*mv(Matrix m, Vector v, Vector vres)
      //	Right multiplies a matrix m by a (column) vector v and stores 
      //	the result in vres.
      //	Returns vres.
      //
      // multvec*vm(Vector v, Matrix m, Vector vres)
      //	Left multiplies a (row) vector v by a matrix m and stores the
      //	result in vres.
      //	Returns vres.

      int * multvec2mv(Matrix2i m, Vector2i v, Vector2i vres);
      float * multvec2mv(Matrix2f m, Vector2f v, Vector2f vres);
      double * multvec2mv(Matrix2d m, Vector2d v, Vector2d vres);

      int * multvec3mv(Matrix3i m, Vector3i v, Vector3i vres);
      float * multvec3mv(Matrix3f m, Vector3f v, Vector3f vres);
      double * multvec3mv(Matrix3d m, Vector3d v, Vector3d vres);

      int * multvec4mv(Matrix4i m, Vector4i v, Vector4i vres);
      float * multvec4mv(Matrix4f m, Vector4f v, Vector4f vres);
      double * multvec4mv(Matrix4d m, Vector4d v, Vector4d vres);

      int * multvec2vm(Vector2i v, Matrix2i m, Vector2i vres);
      float * multvec2vm(Vector2f v, Matrix2f m, Vector2f vres);
      double * multvec2vm(Vector2d v, Matrix2d m, Vector2d vres);

      int * multvec3vm(Vector3i v, Matrix3i m, Vector3i vres);
      float * multvec3vm(Vector3f v, Matrix3f m, Vector3f vres);
      double * multvec3vm(Vector3d v, Matrix3d m, Vector3d vres);

      int * multvec4vm(Vector4i v, Matrix4i m, Vector4i vres);
      float * multvec4vm(Vector4f v, Matrix4f m, Vector4f vres);
      double * multvec4vm(Vector4d v, Matrix4d m, Vector4d vres);

      //////////////////////////////////////////////////////////////////////////
      // multmat*(Matrix m1, Matrix m2, Matrix mres)
      //	Multiply two matrices together and return the result in mres.
      //	returns mres.

      int * multmat2(Matrix2i m1, Matrix2i m2, Matrix2i mres);
      float * multmat2(Matrix2f m1, Matrix2f m2, Matrix2f mres);
      double * multmat2(Matrix2d m1, Matrix2d m2, Matrix2d mres);

      int * multmat3(Matrix3i m1, Matrix3i m2, Matrix3i mres);
      float * multmat3(Matrix3f m1, Matrix3f m2, Matrix3f mres);
      double * multmat3(Matrix3d m1, Matrix3d m2, Matrix3d mres);

      int * multmat4(Matrix4i m1, Matrix4i m2, Matrix4i mres);
      float * multmat4(Matrix4f m1, Matrix4f m2, Matrix4f mres);
      double * multmat4(Matrix4d m1, Matrix4d m2, Matrix4d mres);

      //////////////////////////////////////////////////////////////////////////
      // transpose*(Matrix m, Matrix mres)
      //	Compute the transpose of a matrix and store the result in mres.
      //	Returns mres.

      int * transpose2(Matrix2i m, Matrix2i mres);
      float * transpose2(Matrix2f m, Matrix2f mres);
      double * transpose2(Matrix2d m, Matrix2d mres);

      int * transpose3(Matrix3i m, Matrix3i mres);
      float * transpose3(Matrix3f m, Matrix3f mres);
      double * transpose3(Matrix3d m, Matrix3d mres);

      int * transpose4(Matrix4i m, Matrix4i mres);
      float * transpose4(Matrix4f m, Matrix4f mres);
      double * transpose4(Matrix4d m, Matrix4d mres);

      } // namespace Matrix

   } // namespace arda
#endif // MATRIX_H_
